#include "window.h"
#include "aliases/aliases.h"
#include <SDL2/SDL_rect.h>
#include <SDL2/SDL_render.h>
#include <SDL2/SDL_video.h>
#include <stdbool.h>

bool init_window(window *wnd, const char *title, u32 width, u32 height, i32 x,
                 i32 y) {
  i32 pos_x = x >= 0 ? x : SDL_WINDOWPOS_CENTERED;
  i32 pos_y = y >= 0 ? y : SDL_WINDOWPOS_CENTERED;

  wnd->window =
      SDL_CreateWindow(title, pos_x, pos_y, width, height, SDL_WINDOW_SHOWN);
  if (!(wnd->window)) {
    return false;
  }

  wnd->renderer = SDL_CreateRenderer(
      wnd->window, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
  if (!(wnd->renderer)) {
    cleanup_window(wnd);
    return false;
  }

  wnd->id = SDL_GetWindowID(wnd->window);
  wnd->title = title;

  i32 xmp = -1;
  i32 ymp = -1;
  SDL_GetWindowPosition(wnd->window, &xmp, &ymp);

  wnd->x = xmp;
  wnd->y = ymp;

  i32 wmp = -1;
  i32 hmp = -1;
  SDL_GetWindowSize(wnd->window, &wmp, &hmp);

  wnd->width = wmp;
  wnd->height = hmp;

  return true;
}

void cleanup_window(window *wnd) {
  if (wnd->renderer) {
    SDL_DestroyRenderer(wnd->renderer);
    wnd->renderer = NULL;
  }

  if (wnd->window) {
    SDL_DestroyWindow(wnd->window);
    wnd->window = NULL;
  }

  wnd->width = wnd->height = wnd->id = 0;
}

void set_colour(const window *wnd, colour colour) {
  SDL_SetRenderDrawColor(wnd->renderer, colour.colour.r, colour.colour.g,
                         colour.colour.b, colour.colour.a);
}

void clear_window(const window *wnd, colour colour) {
  set_colour(wnd, colour);
  SDL_RenderClear(wnd->renderer);
}

void swap_buffers(const window *wnd) { SDL_RenderPresent(wnd->renderer); }

void draw_point(const window *wnd, point p, colour colour) {
  set_colour(wnd, colour);
  SDL_RenderDrawPoint(wnd->renderer, p.x, p.y);
}

void draw_line(const window *wnd, const line *ln, colour colour) {
  set_colour(wnd, colour);
  SDL_RenderDrawLine(wnd->renderer, ln->p0.x, ln->p0.y, ln->p1.x, ln->p1.y);
}

void draw_triangle(const window *wnd, triangle triangle, colour colour) {
  line ln0 = {triangle.p0, triangle.p1};
  line ln1 = {triangle.p0, triangle.p2};
  line ln2 = {triangle.p1, triangle.p2};

  draw_line(wnd, &ln0, colour);
  draw_line(wnd, &ln1, colour);
  draw_line(wnd, &ln2, colour);
}

INTERNAL inline i32 min(i32 a, i32 b, i32 c) {
  i32 _min = a <= b ? a : b;
  return _min <= c ? _min : c;
}

INTERNAL inline i32 max(i32 a, i32 b, i32 c) {
  i32 _max = a >= b ? a : b;
  return _max >= c ? _max : c;
}

INTERNAL inline i32 half_space(i32 x1, i32 x2, i32 y1, i32 y2, i32 x, i32 y) {
  return (x2 - x1) * (y - y1) - (y2 - y1) * (x - x1);
}

void fill_triangle(const window *wnd, triangle triangle, colour colour) {
  // Basic triangle filling algorithm from
  // https://web.archive.org/web/20050408192410/http://sw-shader.sourceforge.net/rasterizer.html
  i32 x1 = triangle.p0.x;
  i32 x2 = triangle.p1.x;
  i32 x3 = triangle.p2.x;

  i32 y1 = triangle.p0.y;
  i32 y2 = triangle.p1.y;
  i32 y3 = triangle.p2.y;

  // Find bounding rect
  i32 min_x = min(x1, x2, x3);
  i32 max_x = max(x1, x2, x3);
  i32 min_y = min(y1, y2, y3);
  i32 max_y = max(y1, y2, y3);

  for (i32 y = min_y; y < max_y; ++y) {
    for (i32 x = min_x; x < max_x; ++x) {
      if (half_space(x2, x1, y2, y1, x, y) >= 0 &&
          half_space(x3, x2, y3, y2, x, y) >= 0 &&
          half_space(x1, x3, y1, y3, x, y) >= 0) {
        draw_point(wnd, (point){x, y}, colour);
      }
    }
  }
}

void draw_quad(const window *wnd, quad qd, colour colour) {
  line l0 = (line){qd.p0, qd.p1};
  line l1 = (line){qd.p1, qd.p3};
  line l2 = (line){qd.p3, qd.p2};
  line l3 = (line){qd.p2, qd.p0};

  draw_line(wnd, &l0, colour);
  draw_line(wnd, &l1, colour);
  draw_line(wnd, &l2, colour);
  draw_line(wnd, &l3, colour);
}

void fill_quad(const window *wnd, quad qd, colour colour) {
  triangle t0 = (triangle){qd.p3, qd.p1, qd.p0};
  triangle t1 = (triangle){qd.p0, qd.p2, qd.p3};

  fill_triangle(wnd, t0, colour);
  fill_triangle(wnd, t1, colour);
}

void draw_rect(const window *wnd, rect rec, colour colour) {
  set_colour(wnd, colour);

  SDL_Rect dst = {rec.topleft.x, rec.topleft.y, rec.w, rec.h};

  SDL_RenderDrawRect(wnd->renderer, &dst);
}

void fill_rect(const window *wnd, rect rec, colour colour) {
  set_colour(wnd, colour);

  SDL_Rect dst = {rec.topleft.x, rec.topleft.y, rec.w, rec.h};

  SDL_RenderFillRect(wnd->renderer, &dst);
}
